package org.ananas.hc.compiler;

import java.io.File;
import java.util.List;
import java.util.ArrayList;
import com.sun.javadoc.Tag;
import com.sun.javadoc.RootDoc;
import com.sun.javadoc.ClassDoc;
import com.sun.javadoc.Parameter;
import com.sun.javadoc.MethodDoc;
import org.ananas.hc.MessageStore;
import com.sun.javadoc.DocErrorReporter;
import org.xml.sax.helpers.NamespaceSupport;

public class CompilerDoclet
{
   public static class DocletMessenger
      implements Messenger
   {
      protected DocErrorReporter reporter;

      public DocletMessenger(DocErrorReporter reporter)
      {
         this.reporter = reporter;
      }

      public void error(CompilerException e)
         throws CompilerException
      {
         reporter.printError(e.getLocalizedMessage());
      }

      public void fatal(CompilerException e)
         throws CompilerException
      {
         reporter.printError(e.getLocalizedMessage());
         throw e;
      }

      public void warning(CompilerException e)
      {
         info(e.getLocalizedMessage());
      }

      public void info(String msg)
      {
         reporter.printNotice(msg);
      }
   }

   private static MessageStore getMessageStore()
   {
      return MessageStore.getMessageStore("org.ananas.hc.compiler.Message");
   }

   public static boolean start(RootDoc root)
      throws Exception
   {
      try
      {
         String[][] options = root.options();
         File destdir = new File(".");
         for(int i = 0;i < options.length;i++)
            if(options[i][0].equals("-d"))
               destdir = new File(options[i][1]);
         CompilerDoclet compiler = new CompilerDoclet();
         HandlerInfo[] handlers = compiler.compile(root);
         Messenger messenger = new DocletMessenger(root);
         HCTablesGenerator generator =
            new HCTablesGenerator(getMessageStore(),messenger,destdir);
         for(int i = 0;i < handlers.length;i++)
            generator.generateHCTables(handlers[i]);
         return true;
      }
      catch(CompilerException e)
      {
         // no need to display again, it has already been shown
         // to the user
         return false;
      }
   }

   public static boolean validOptions(String[][] options,
                                      DocErrorReporter error)
   {
      for(int i = 0;i < options.length;i++)
         if(options[i][0].equals("-d"))
         {
            File file = new File(options[i][1]);
            if(!file.exists() || !file.isDirectory())
            {
               MessageStore message = getMessageStore();
               error.printError(message.getMessage("InvalidDestinationDirectory",
                                                   options[i][1]));
               return false;
            }
         }
      return true;
   }

   public static int optionLength(String option)
   {
      if(option.equals("-d"))
         return 2;
      return 0;
   }

   protected HandlerInfo[] compile(RootDoc root)
   {
      ClassDoc[] classes = root.classes();
      List handlers = new ArrayList();
      for(int i = 0;i < classes.length;i++)
      {
         HandlerInfo handler = compile(classes[i]);
         if(handler != null)
            handlers.add(handler);
      }
      HandlerInfo[] handlersArray = new HandlerInfo[handlers.size()];
      handlers.toArray(handlersArray);
      return handlersArray;
   }

   protected HandlerInfo compile(ClassDoc clasz)
   {
      ClassDoc[] interfaces = clasz.interfaces();
      if(interfaces == null)
         return null;
      boolean found = false;
      for(int i = 0;i < interfaces.length;i++)
         if(interfaces[i].qualifiedName().equals("org.ananas.hc.HCHandler"))
            found = true;
      if(!found)
         return null;

      Tag[] tags = clasz.tags("xmlns");
      NamespaceSupport namespaceSupport = new NamespaceSupport();
      if(tags != null)
         for(int i = 0;i < tags.length;i++)
         {
            String content = tags[i].text();
            int pos = content.indexOf(' ');
            if(pos == -1)
               namespaceSupport.declarePrefix("",content);
            else
               namespaceSupport.declarePrefix(content.substring(0,pos),
                                              content.substring(pos + 1));
         }

      MethodDoc[] methods = clasz.methods();
      List methodsList = new ArrayList();
      if(methods != null)
         for(int i = 0;i < methods.length;i++)
         {
            MethodInfo method = compile(methods[i]);
            if(method != null)
               methodsList.add(method);
         }
      MethodInfo[] methodsArray = new MethodInfo[methodsList.size()];
      methodsList.toArray(methodsArray);

      return new HandlerInfo(clasz.qualifiedName(),
                             namespaceSupport,
                             methodsArray);
   }

   protected MethodInfo compile(MethodDoc method)
   {
      Tag[] tags = method.tags("xpath");
      if(tags == null || tags.length <= 0)
         return null;
      List xpathList = new ArrayList();
      for(int i = 0;i < tags.length;i++)
         xpathList.add(tags[i].text());
      String[] xpaths = new String[xpathList.size()];
      xpathList.toArray(xpaths);

      Parameter[] params = method.parameters();
      List paramsList = new ArrayList();
      for(int i = 0;i < params.length;i++)
      {
         String[] p = new String[2];
         p[0] = params[i].typeName();
         p[1] = params[i].name();
         paramsList.add(p);
      }
      String[][] paramsArray = new String[paramsList.size()][];
      paramsList.toArray(paramsArray);

      return new MethodInfo(method.name(),
                            xpaths,
                            paramsArray);
   }
}
