package org.ananas.hc.compiler;

import java.util.Set;
import java.util.Vector;
import java.util.HashSet;
import java.util.Iterator;
import org.ananas.hc.QName;
import java.util.Enumeration;

public class DFAFactory
{
   public QName[] uniqueSymbols(XPathNode root)
   {
      Set qnames = new HashSet();
      uniqueSymbolsHelper(root,qnames);
      QName[] array = new QName[qnames.size()];
      return (QName[])qnames.toArray(array);
   }

   public void uniqueSymbolsHelper(XPathNode node,Set qnames)
   {
      if(node == null)
         return;
      QName qname = node.toQName();
      if(qname != null)
         qnames.add(qname);
      uniqueSymbolsHelper(node.getLeft(),qnames);
      uniqueSymbolsHelper(node.getRight(),qnames);
   }

   /**
    * <P>constructs the DFA
    * <P>uses the algorithm 3.5 in "Compilers Principles, Techniques and Tools",
    * Alfred V. Aho, Ravi Sethi, Jeffrey D. Ullman, Addison-Wesley, 1986
    */
   public DFATable createDFA(XPathNode root)
   {
      // the algorithm is used as comments, the algorithm is preceded by -
      Vector dstates = new Vector(),
             dtran = new Vector(),
             astate = new Vector();
      QName[] symbols = uniqueSymbols(root);

      // - initially the only unmarked state in Dstates is is firstpos(root)
      dstates.addElement(root.firstpos());

      Enumeration statesEnumeration = dstates.elements();

      // - while there is an unmarked state T in Dstates do
      while(statesEnumeration.hasMoreElements())
      {
         // - mark T
         // I do not have to explicitely mark items because I always add new
         // ones at the end and I mark them as soon as I reach them (more
         // efficient then searching the vector for unmarked items)
         // the price to pay is that Dstates is not sorted and contains() is
         // O(n)
         Set t = (Set)statesEnumeration.nextElement();

         // create a new column for Dtran (symbol.length + 1 for the joker)
         int dtranOfT[] = new int[symbols.length];
         dtran.addElement(dtranOfT);

         // - for each input symbol a do
         for(int i = 0;i < symbols.length;i++)
         {
            // - let U be the set of positions that are in followpos(p)
            // - for some position p in T,
            // - such that the symbol at position p is a
            Set u = e_closure(symbols[i],t);

            // - if U is not empty and is not in Dstates then
            // - add U as an unmarked state to Dstates
            if (!u.isEmpty() && !dstates.contains(u))
               dstates.addElement(u);

            // - Dtran[T,a] = U;
            // I do not store U but its index in Dstates which is a sort of name
            // for U. It's just like when you have two or three sets and you say
            // "let's call { a, b, c } 1, { a, f } 2, etc.
            // When it goes in the table, it's the same thing.
            dtranOfT[i] = dstates.indexOf(u);
         }
         // do the same thing for the joker character...
         // cf. the for loop for comments
         // I am under the impression we could do the same thing
         // if we were to adapt e_closure() so it would compute
         // the joker with specific rules...
         /*
         {
            Set u = e_closureForJoker(t);
            if (!u.isEmpty() && ! dstates.contains(u))
               dstates.addElement(u);
            dtranOfT[symbolSet.length] = dstates.indexOf(u);
         }
         */

         // if the state is not an accepting state, it return null
         // so this adds a null value for non-accepting states and
         // the data object for accepting ones
         astate.addElement(getAcceptingState(t));
      }

      int[][] idtran = new int[dtran.size()][symbols.length];
      dtran.copyInto(idtran);
      Object[] oastate = new Object[astate.size()];
      astate.copyInto(oastate);
      DFATable table = new DFATable(symbols,idtran,oastate);
      return table;
   }

   private Object getAcceptingState(Set t)
   {
      Iterator iterator = t.iterator();
      Object data = null;
      int priority = Integer.MIN_VALUE;
      while(iterator.hasNext())
      {
         XPathNode n = (XPathNode)iterator.next();
         if(n.getType() == XPathNode.END_MARK &&
             priority < n.getPriority())
         {
            priority = n.getPriority();
            data = n.getData();
         }
      }
      return data;
   }

   /**
    * <P>helper function for constructDFA(), it constructs U, the epsilon
    * closure as the set of positions that are in followpos(p) for some position
    * p in T, such that the symbol at position p is a
    * <P>i.e. the subset of t for the elements of t that represents a
    */
   private Set e_closure(QName a,Set t)
   {
      Set result = new HashSet();
      Iterator iterator = t.iterator();
      while(iterator.hasNext())
      {
         XPathNode n = (XPathNode)iterator.next();
         if(a.matches(n))
            result.addAll(n.followpos());
      }
      return result;
   }
}
