package org.ananas.hc.compiler;

import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import org.ananas.hc.QName;
import java.text.MessageFormat;
import org.ananas.hc.NotImplementedException;

public class XPathNode
   extends QName
{
   public final static int OR_XPATH = LAST_QNAME_TYPE + 1,
                           STAR = LAST_QNAME_TYPE + 2,
                           PARENT_OF = LAST_QNAME_TYPE + 3;

   private XPathNode left,
                     right;
   private boolean nullable;
   private Set firstpos,
               lastpos,
               followpos = null;
   private int priority;
   private Object data;

   public XPathNode(int type,
                    String namespaceURI,
                    String localName,
                    XPathNode left,
                    XPathNode right,
                    int priority,
                    Object data)
   {
      super(type,namespaceURI,localName);
      this.left = left;
      this.right = right;
      if(data == null)
         this.priority = Integer.MIN_VALUE;
      else
         this.priority = priority;
      this.data = data;
      followpos = new HashSet();
      switch(type)
      {
         case ROOT:
         case ELEMENT:
         case ATTRIBUTE:
         case JOKER:
         case END_MARK:
            nullable = false;
            firstpos = new HashSet();
            firstpos.add(this);
            lastpos = new HashSet();
            lastpos.add(this);
            break;
         case PARENT_OF:
            nullable = left.nullable() && right.nullable();
            firstpos = new HashSet(left.firstpos());
            if(left.nullable())
               firstpos.addAll(right.firstpos());
            lastpos = new HashSet(right.lastpos());
            if(right.nullable())
               lastpos.addAll(left.lastpos());
            Iterator iterator = left.lastpos().iterator();
            while(iterator.hasNext())
            {
               XPathNode n = (XPathNode)iterator.next();
               n.followpos.addAll(right.firstpos());
            }
            break;
         case OR_XPATH:
            nullable = left.nullable() || right.nullable();
            firstpos = new HashSet(left.firstpos());
            firstpos.addAll(right.firstpos());
            lastpos = new HashSet(left.lastpos());
            lastpos.addAll(right.lastpos());
            break;
         case STAR:
            nullable = true;
            firstpos = new HashSet(left.firstpos());
            lastpos = new HashSet(left.lastpos());
            iterator = left.lastpos().iterator();
            while(iterator.hasNext())
            {
               XPathNode n = (XPathNode)iterator.next();
               n.followpos.addAll(firstpos);
            }
            break;
         default:
            throw new NotImplementedException("unknown type");
      }
   }

   public XPathNode(int type,
                    String namespaceURI,
                    String localName,
                    XPathNode left,
                    XPathNode right)
   {
      this(type,namespaceURI,localName,
           left,right,Integer.MIN_VALUE,null);
   }

   public XPathNode(int type)
   {
      this(type,null,null,null,null,Integer.MIN_VALUE,null);
   }

   public XPathNode(int type,int priority,Object data)
   {
      this(type,null,null,null,null,priority,data);
   }

   public XPathNode(int type,
                    String namespaceURI,
                    String localName)
   {
      this(type,namespaceURI,localName,
           null,null,Integer.MIN_VALUE,null);
   }

   public XPathNode(int type,XPathNode left,XPathNode right)
   {
      this(type,null,null,left,right,Integer.MIN_VALUE,null);
   }

   public XPathNode(int type,XPathNode left)
   {
      this(type,null,null,left,null,Integer.MIN_VALUE,null);
   }

   protected void setType(int type)
   {
      if(type < ROOT || type > PARENT_OF)
         throw new IllegalArgumentException("unknown type");
      else
         this.type = type;
   }

   public boolean equals(Object object)
   {
      return object == this;
   }

   public int hashCode()
   {
      return System.identityHashCode(this);
   }

   public boolean isSynonymousWith(XPathNode node,boolean deep)
   {
      if(!super.equals(node))
         return false;
      if(deep)
      {
         if(left == null)
            if(right == null)
               return node.left == null && node.right == null;
            else
               return node.left == null &&
                      right.isSynonymousWith(node.right,true);
         else
            if(right == null)
               return left.isSynonymousWith(node.left,true) &&
                      node.right == null;
            else
               return left.isSynonymousWith(node.left,true) &&
                      right.isSynonymousWith(node.right,true);
      }
      else
         return left == node.left && right == node.right;
   }

   public XPathNode getLeft()
   {
      return left;
   }

   public XPathNode getRight()
   {
      return right;
   }

   public int getPriority()
   {
      return priority;
   }

   public Object getData()
   {
      return data;
   }

   public boolean nullable()
   {
      return nullable;
   }

   public Set firstpos()
   {
      return firstpos;
   }

   public Set lastpos()
   {
      return lastpos;
   }

   public Set followpos()
   {
      return followpos;
   }

   public String toString()
   {
      String pattern = null;
      switch(type)
      {
         case ROOT:
         case ATTRIBUTE:
         case ELEMENT:
            return super.toString();
         case END_MARK:
            if(data == null)
               return "(#)";
            else
               return MessageFormat.format("(#), {0}",
                                           new Object[] { data.toString() });
         case OR_XPATH:
            pattern = "({0}) | ({1})";
            break;
         case PARENT_OF:
            pattern = "{0} / {1}";
            break;
         default:
            throw new NotImplementedException("unknown type");
      }
      return MessageFormat.format(pattern,
                                  new Object[] { left.toString(),
                                                 right.toString() });
   }

   public boolean isQName()
   {
      return type <= LAST_QNAME_TYPE;
   }

   public QName toQName()
   {
      if(isQName())
         return new QName(type,namespaceURI,localName);
      else
         return null;
   }
}
